//==========================================================================
// SASImportML.cpp
//
// Institution: Department of Mathematics
// 			    University of North Carolina
// Date: 12-04-2010
// Copyright: None.
//
// SASImport.exe is a sample Windows program which provides a function 
// (named "SASImport") you can call from within a Mathematica notebook 
// to import a SAS data set.  SASImport uses SAS Integration Technologies 
// to open the SAS data set and MathLink to communicate with Mathematica.
//
// To use, execute the following commands from within your Mathematica notebook:
//
//		mlink = Install["<path to SASImport.exe goes here>"]
//		sasdata= SASImport["<path to SAS Data Set goes here>"]
//		Uninstall[mlink]
//
// See the "SASImport Example.nb" sample notebook for more examples.
//
// This source code is not copyrighted.  You are free to use, copy,
// and/or modify this source code however you like.
// Note that this source code is provided "as is" with no warranties
// of any kind implied whatsoever.
//
// For instructions on how to create a Visual Studio project in order
// to build SASImport.exe, consult the file: SASImport ReadMe.txt
//
// SAS(c) is a product of SAS Institute, Inc.
//==========================================================================
#include "stdafx.h"
#include "mathlink.h"



#import "c:\program files\common files\system\ado\msado15.dll" rename("EOF", "adoEOF")

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//		Interface between SASImport and MathLink
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#ifdef __cplusplus
 extern "C" {
#endif 

extern void sasImport( const char * szPath );

#ifdef __cplusplus
 }
#endif

#ifdef __cplusplus
 extern "C" {
#endif

static void importSASDataSetToMathematica( const char *szPath );



//This routine is invoked when the user executes the SASImport[s] command in his/her Mathematica notebook.
void  sasImport( const char * szPath ) {
    importSASDataSetToMathematica( szPath );
}

#ifdef __cplusplus
 }
#endif


#if WINDOWS_MATHLINK

#if __BORLANDC__
#pragma argsused
#endif


//Note: When we are launched by Mathematica, this WinMain will be executed.
//Do NOT place this initialization code in our InitInstance() routine.
//Note that we have customized this version of WinMain to handle initializing
//and uninitializing OLE.
int PASCAL WinMain( HINSTANCE hinstCurrent, HINSTANCE hinstPrevious, LPSTR lpszCmdLine, int nCmdShow)
{
	char  buff[512];
	char FAR * buff_start = buff;
	char FAR * argv[32];
	char FAR * FAR * argv_end = argv + 32;
    int         result_MLMain;

	hinstPrevious = hinstPrevious; /* suppress warning */

	if( !MLInitializeIcon( hinstCurrent, nCmdShow)) return 1;
	MLScanString( argv, &argv_end, &lpszCmdLine, &buff_start);

#if _WIN32_WINNT >= 0x0400 & defined(_ATL_FREE_THREADED)
    HRESULT hRes = CoInitializeEx(NULL, COINIT_MULTITHREADED);
#else
    HRESULT hRes = CoInitialize(NULL);
#endif

    result_MLMain = MLMain( (int)(argv_end - argv), argv);


    //AfxMessageBox("MLMain returned");

    CoUninitialize();

	return result_MLMain;
}

#else

int main(int argc, char* argv[])
{
	return MLMain(argc, argv);
}

#endif

//this enum corresponds to the list of messages defined in SASImport.tm
enum SASImportError {
	eSASImportErrorFileDNE = MLFirstUserMessage+1,
	eSASImportErrorUnableToConnect,
	eSASImportErrorUnrecognizedFileType,
	eSASImportErrorUnableToParsePath,
	eSASImportErrorUnableToOpenSASDS,
	eSASImportErrorUnableToImportSASDS,
	eSASImportErrorUnableToOpenFile,
	eSASImportErrorUnableToAllocMem,
	eSASImportErrorUnknownError
};

static void sendErrorToMathematica( SASImportError nWhichError, char* szOptionalMsg=NULL );

static void importSASDataSetToMathematica( const char *szPath ) {
	//make sure the file exists
	CFileStatus status;
	if( CFile::GetStatus( szPath, status )==0 ) {
		sendErrorToMathematica(eSASImportErrorFileDNE);
		return;
	}
	CString		cstrPath(szPath);
	BOOL	bIsSASDataSet = (cstrPath.GetLength() >= 10 &&
							cstrPath.Right(9).MakeLower() == ".sas7bdat") ? TRUE : FALSE;
	if ( !bIsSASDataSet ) {
		sendErrorToMathematica(eSASImportErrorUnrecognizedFileType);
		return;
	}

	int		ndx = cstrPath.ReverseFind('\\');
	if (ndx==-1) {
		sendErrorToMathematica(eSASImportErrorUnableToParsePath);
		return;
	}
	CString		cstrDataSource;
	cstrDataSource = cstrPath.Left(ndx);

	CString		cstrDataSetName = cstrPath.Left(cstrPath.GetLength()-9);//chop off ".sas7bdat" extension
	ndx++;	//skip over the final slash
	cstrDataSetName = cstrDataSetName.Right(cstrDataSetName.GetLength()-ndx);
	
	//CString	cstrDebugMsg("DataSource=");
	//cstrDebugMsg += cstrDataSource;
	//cstrDebugMsg += " Data set=";
	//cstrDebugMsg += cstrDataSetName;
	//AfxMessageBox(cstrDebugMsg);

	ADODB::_ConnectionPtr obConnection = NULL;
	ADODB::_RecordsetPtr   obRecordset = NULL;	
	HRESULT	hr = obRecordset.CreateInstance(  __uuidof( ADODB::Recordset  ));
	hr = obConnection.CreateInstance(  __uuidof( ADODB::Connection  ));

	CString	cstrConnect;
	cstrConnect = "Provider=SAS.LocalProvider.1; Data Source=";
	cstrConnect += cstrDataSource;

	//_bstr_t strConnect( "Provider=SAS.LocalProvider.1; Data Source=C:\\SASImport\\sampleSASdata" );
	_bstr_t strConnect( (LPCSTR)cstrConnect );
	try {
		hr = obConnection->Open(strConnect, _bstr_t(""), _bstr_t(""), NULL);
	}
	catch(_com_error e){
		//CString		errMsg("SASImport Error: Unable to connect to SAS.LocalProvider.1 - ");
		//errMsg += (char *)e.Description();
		//AfxMessageBox(errMsg);
		sendErrorToMathematica(eSASImportErrorUnableToConnect);
		return;
	}


	//_bstr_t strSource = "sta6207.sas7bdat";
	_bstr_t strSource = (LPCSTR)cstrDataSetName;

	try {
		obRecordset->Open(strSource, _variant_t((IDispatch *)obConnection),
			ADODB::adOpenStatic /* ADODB::adOpenForwardOnly */ ,
			ADODB::adLockBatchOptimistic /* ADODB::adLockReadOnly */ , 
			ADODB::adCmdTableDirect );
		if ( obRecordset==NULL )
			throw(eSASImportErrorUnableToOpenSASDS);

		long fieldCount = obRecordset->Fields->Count;
        long recCount   = obRecordset->RecordCount;

		//char    szBuf[256];
        //sprintf(szBuf,"fieldCount=%d recCount=%d", fieldCount, recCount );
		//AfxMessageBox(szBuf);

        ADODB::FieldsPtr		pFldLoop = obRecordset->GetFields();
        ADODB::PropertiesPtr    pPrpLoop = obRecordset->GetProperties();

		if ( recCount==0 || fieldCount==0 ) {//return an empty List to Mathematica
			MLPutFunction( stdlink, "List", 0 );
			MLEndPacket( stdlink);
			return;
		}

		//send a new list to Mathematica to encapsulate the records in the data set
		if ( MLPutFunction( stdlink, "List", recCount+1 )==0 ) {//add 1 to account for list of field names
			throw(eSASImportErrorUnableToImportSASDS);
		}

		_variant_t vtIndex;
        vtIndex.vt = VT_I2;

		//send a new list to Mathematica to encapsulate the set of field names
		if ( MLPutFunction( stdlink, "List", fieldCount )==0 ) {
			//AfxMessageBox("Unable to create new List");
			throw(eSASImportErrorUnableToImportSASDS);
		}
		//send the list of field names
		for( int nFldNo=0; nFldNo < fieldCount; nFldNo++ ) {
			vtIndex.iVal = nFldNo;
			_bstr_t colName = pFldLoop->GetItem(vtIndex)->Name;
			MLPutString( stdlink, (LPCSTR)colName );
		}

		//AfxMessageBox("Sent column names");
		//CString	cstrDebugMsg;
		//char	szRecNoBuf[256];

		//now send the data records to Mathematica
        for( int nRecordNo=0; nRecordNo < recCount; nRecordNo++ ) {
			//send a new list to Mathematica for the current row of data
            if ( MLPutFunction( stdlink, "List", fieldCount )==0 ) {
                throw(eSASImportErrorUnableToImportSASDS);
            }

			BOOL	bMissingValue=FALSE;
            for( int nFldNo=0; nFldNo < fieldCount; nFldNo++ ) {

                vtIndex.iVal = nFldNo;

                _bstr_t colName = pFldLoop->GetItem(vtIndex)->Name;
                //OutputDebugString("Column: " );
                //OutputDebugString((LPCSTR)colName);

                //http://msdn.microsoft.com/en-us/library/ms675318(VS.85).aspx
                ADODB::DataTypeEnum dtype = pFldLoop->GetItem(vtIndex)->Type;
                
                switch( dtype ) {
                case ADODB::adVarWChar:
                case ADODB::adWChar:
                case ADODB::adVarChar:
                case ADODB::adChar:
                    {
						//check for missing value
						_variant_t fldVal = pFldLoop->GetItem(vtIndex)->Value;
						if ( fldVal.vt==VT_NULL || fldVal.vt==VT_EMPTY ) { //missing value
							MLPutFunction( stdlink, "Missing",0);
							//AfxMessageBox("Value is missing");
							break;
						}
                        //OutputDebugString(" - Type is char - Value is ");
                        _bstr_t valStr = (_bstr_t)fldVal;
						if (valStr.length()==0) {//missing value
							MLPutFunction( stdlink, "Missing",0);
							//AfxMessageBox("Value is missing");
							break;
						}
						MLPutString( stdlink, (LPCSTR)valStr );
                        //OutputDebugString((LPCSTR)valStr);
                    }
                    break;
                //FixMe Incomplete! Cf url above for complete list of possible dtypes
                case ADODB::adInteger:
				case ADODB::adDouble:
                    {
                        //OutputDebugString(" - Type is numeric ");
						//AfxMessageBox("Type is numeric");

                        _variant_t fldVal = pFldLoop->GetItem(vtIndex)->Value;

						//check for missing value
						if ( fldVal.vt==VT_NULL ) { //missing value
							MLPutFunction( stdlink, "Missing",0);
							//AfxMessageBox("Value is missing");
							break;
						}
						
                        double  dVal = 0.;
                        switch( dtype ) {
                        case ADODB::adDouble:
                            dVal = (double)fldVal.dblVal;
                            break;
                        case ADODB::adInteger:
                            dVal = (double)fldVal.intVal;
                            break;
                        }
                        //char    szBuf[256];
                        //sprintf(szBuf,"%f", dVal);
                        //OutputDebugString(szBuf);
						MLPutDouble( stdlink, dVal );
                    }
                    break;
				//According to sas.support, ADO represents missing numeric values as a Variant data type with a value of Null
				//However, we don't ever seem to hit this case when we have missing values
				case ADODB::adVariant:
					MLPutFunction( stdlink, "Missing",0);
					break;
                default:
					MLPutFunction( stdlink, "Missing",0);
                    break;
                }
                //OutputDebugString("\n");
                
                
            }//for each field in current record

			//cstrDebugMsg = "Sent record #";
			//sprintf(szRecNoBuf,"%d", nRecordNo);
			//cstrDebugMsg += szRecNoBuf;
			//AfxMessageBox(cstrDebugMsg);

            obRecordset->MoveNext();
        }//for each record

        obRecordset->Close();
    }
	catch( SASImportError errCode ) {
		sendErrorToMathematica( errCode );
	}
	catch(_com_error e){
		sendErrorToMathematica(eSASImportErrorUnableToOpenSASDS, (char *)e.Description() );
	}
	obConnection->Close();

	MLEndPacket( stdlink); //close off the list representing the data set
    MLFlush( stdlink );
}

//In case we encounter an error, we call this routine to send a message back to
//Mathematica.  The text of the errror messages is provided in our SASImport.tm file.
static void sendErrorToMathematica( SASImportError nWhichErrorIn, char* szOptionalMsg ) {
	CString		cstrWhichError;
	switch( nWhichErrorIn ) {
	case eSASImportErrorFileDNE:
		cstrWhichError = "fileDNE";
		break;
	case eSASImportErrorUnableToConnect:
		cstrWhichError = "unableToConnect";
		break;
	case eSASImportErrorUnrecognizedFileType:
		cstrWhichError = "unrecognizedFileType";
		break;
	case eSASImportErrorUnableToParsePath:
		cstrWhichError = "unableToParsePath";
		break;
	case eSASImportErrorUnableToOpenSASDS:
		cstrWhichError = "unableToOpenSASDS";
		break;
	case eSASImportErrorUnableToImportSASDS:
		cstrWhichError = "unableToImportSASDS";
		break;
	case eSASImportErrorUnableToOpenFile:
		cstrWhichError = "unableToOpenFile";
		break;
	case eSASImportErrorUnableToAllocMem:
		cstrWhichError = "unableToAllocMem";
		break;
	default:
		cstrWhichError = "unknownError";
		break;
	}
	CString		cstrError;
	cstrError.Format("Message[SASImport::%s]", cstrWhichError );
	if ( szOptionalMsg!=NULL )
		cstrError += szOptionalMsg;
	MLEvaluate(stdlink, cstrError.GetBuffer() );
	MLNextPacket(stdlink);
	MLNewPacket(stdlink);
	MLPutSymbol(stdlink, "$Failed");
}
